/* Results from 1 million rows 
Strong typing had a query time of 20 seconds and uses 71MB disk space
Weak typing had a query time of 56 seconds and uses 131MB disk space

The reason for the larger disk space is due to the untyped column for 
storing attribute values. In the example '1000000' uses 7 bytes, whereas
the integer 1000000 only uses 4 bytes. Further, when using strong typing
even surrogate keys can be optimized, as with the state attribute which 
only uses a bit as a key.

When it comes to query time, the deciding factor is that type conversions
have to be done to bring back the correct data type, as well as the larger
amount of data that has to be scanned for the same amount of rows. Note 
however that if for example the value 99 occurs many times, the weak typed
model will only contain it once, whereas the anchor model will contain it
as many times as it occurs. To resolve this in the anchor model, in the 
cases where values are predictable and frequent, they should be modeled 
as domain value knots.
*/

use master;
go
-- table to store results
IF EXISTS (
	SELECT 
		*
    FROM 
		tempdb..sysobjects
    WHERE 
		type = 'U' 
	and 
		NAME like '#timers%'
) 
drop table #timers;

create table #timers (
	test varchar(100),
	startTime datetime,
	stopTime datetime,
	duration as datediff(ms, startTime, stopTime)
);

-- *************** STRONG TYPING ***************
IF  EXISTS (SELECT name FROM sys.databases WHERE name = N'StrongTyping')
DROP DATABASE [StrongTyping];
go
create database StrongTyping ON  PRIMARY (
	NAME = N'StrongTyping', 
	FILENAME = N'C:\Program\Microsoft SQL Server\MSSQL.1\MSSQL\DATA\StrongTyping.mdf',
	SIZE = 1024000KB, 
	MAXSIZE = UNLIMITED, 
	FILEGROWTH = 10%
);
go
ALTER DATABASE StrongTyping SET RECOVERY SIMPLE 
GO
use StrongTyping;

-- initialization parameters
declare @rows int;
set @rows = 1000000;

create table AN_Anchor (
	AN_ID int not null,
	_metadata int not null,
	primary key (
		AN_ID asc
	)
);

create table ANSTA_AnchorState (
	AN_ID int not null,
	STA_ID bit not null,
	ANSTA_FromDate datetime not null,
	_metadata int not null,
	primary key (
		AN_ID asc,
		ANSTA_FromDate desc
	)
);

create table ANVAL_AnchorValue (
	AN_ID int not null,
	ANVAL_Value int not null,
	ANVAL_FromDate datetime not null,
	_metadata int not null,
	primary key (
		AN_ID asc,
		ANVAL_FromDate desc
	)
);

create table STA_State (
	STA_ID bit not null,
	STA_Name char(3) not null,
	primary key (
		STA_ID asc
	)
);

WITH gen(num) AS (
	SELECT
		1
	UNION ALL
	SELECT
		num + 1
	FROM
		gen
	WHERE
		num < @rows
)
insert into AN_Anchor
select
	num,
	1
from
	gen
option (
	maxrecursion 0
);

WITH gen(num) AS (
	SELECT
		1
	UNION ALL
	SELECT
		num + 1
	FROM
		gen
	WHERE
		num < @rows
)
insert into ANSTA_AnchorState
select
	num,
	num % 2,
	'2008-01-01',
	1
from
	gen
option (
	maxrecursion 0
);

insert into STA_State
values (
	0,
	'Off'
);

insert into STA_State
values (
	1,
	'On'
);

WITH gen(num) AS (
	SELECT
		1
	UNION ALL
	SELECT
		num + 1
	FROM
		gen
	WHERE
		num < @rows
)
insert into ANVAL_AnchorValue
select
	num,
	num,
	'2008-01-01',
	1
from
	gen
option (
	maxrecursion 0
);

go
create view lAN_Anchor
as
select
	an.AN_ID,
	sta.STA_Name,
	anval.ANVAL_Value
from
	AN_Anchor an
left join
	ANSTA_AnchorState ansta
on
	ansta.AN_ID = an.AN_ID
and
	ansta.ANSTA_FromDate = (
		select
			max(sub.ANSTA_FromDate)
		from
			ANSTA_AnchorState sub
		where
			sub.AN_ID = an.AN_ID
	)
left join
	STA_State sta
on
	sta.STA_ID = ANSTA.STA_ID
left join
	ANVAL_AnchorValue anval
on
	anval.AN_ID = an.AN_ID
and
	anval.ANVAL_FromDate = (
		select
			max(sub.ANVAL_FromDate)
		from
			ANVAL_AnchorValue sub
		where
			sub.AN_ID = an.AN_ID
	);

go
-- testing
DBCC FREEPROCCACHE;
DBCC DROPCLEANBUFFERS;

insert into #timers (test, startTime) 
values(
	'Strong typing', 
	GETDATE()
);

select 
	STA_Name,
	max(ANVAL_Value)
from
	lAN_Anchor
group by
	STA_Name;

update #timers 
set 
	stopTime = GETDATE() 
where 
	test = 'Strong typing';

go
use master;
go
-- *************** WEAK TYPING ***************
IF  EXISTS (SELECT name FROM sys.databases WHERE name = N'WeakTyping')
DROP DATABASE [WeakTyping];
go
create database WeakTyping ON  PRIMARY (
	NAME = N'WeakTyping', 
	FILENAME = N'C:\Program\Microsoft SQL Server\MSSQL.1\MSSQL\DATA\WeakTyping.mdf',
	SIZE = 1024000KB, 
	MAXSIZE = UNLIMITED, 
	FILEGROWTH = 10%
);
go
ALTER DATABASE WeakTyping SET RECOVERY SIMPLE 
GO
use WeakTyping;

-- initialization parameters
declare @rows int;
set @rows = 1000000;

create table SomeTable (
	SomeTable_ID int not null,
	_metadata int not null,
	primary key (
		SomeTable_ID asc
	)
);

create table GenericAttribute (
	GenericAttribute_ID bigint identity(1, 1) not null,
	GenericAttribute_Value varchar(255) not null,
	GenericAttribute_Type int not null,
	_metadata int not null,
	primary key (
		GenericAttribute_ID asc
	)
);

insert into GenericAttribute (
	GenericAttribute_Value,
	GenericAttribute_Type,
	_metadata
)
values (
	'On',
	1,
	1
);

insert into GenericAttribute (
	GenericAttribute_Value,
	GenericAttribute_Type,
	_metadata
)
values (
	'Off',
	1,
	1
);

WITH gen(num) AS (
	SELECT
		1
	UNION ALL
	SELECT
		num + 1
	FROM
		gen
	WHERE
		num < @rows
)
insert into GenericAttribute (
	GenericAttribute_Value,
	GenericAttribute_Type,
	_metadata
)
select
	cast(num as varchar(255)),
	2,
	1
from
	gen
option (
	maxrecursion 0
);

create table AttributeType (
	AttributeType_ID int not null,
	AttributeType_Name varchar(255) not null,
	AttributeType_DataType varchar(255) not null,
	_metadata int not null,
	primary key (
		AttributeType_ID asc
	)
);

insert into AttributeType
values (
	1,
	'State',
	'char(3)',
	1
);

insert into AttributeType
values (
	2,
	'Value',
	'int',
	1
);

create table SomeTableAttribute (
	SomeTable_ID int not null,
	GenericAttribute_ID int not null,
	SomeTableAttribute_FromDate datetime not null,
	_metadata int not null,
	primary key (
		SomeTable_ID asc,
		GenericAttribute_ID asc,
		SomeTableAttribute_FromDate desc
	)
);

WITH gen(num) AS (
	SELECT
		1
	UNION ALL
	SELECT
		num + 1
	FROM
		gen
	WHERE
		num < @rows
)
insert into SomeTable
select
	num, 
	1
from
	gen
option (
	maxrecursion 0
);

WITH gen(num) AS (
	SELECT
		1
	UNION ALL
	SELECT
		num + 1
	FROM
		gen
	WHERE
		num < @rows
)
insert into SomeTableAttribute 
select
	g.num,
	a.GenericAttribute_ID,
	'2008-01-01',
	1
from
	gen g
join
	GenericAttribute a
on
	case 
		when g.num % 2 = 0 then 'Off'
		when g.num % 2 = 1 then 'On'
		else null
	end = a.GenericAttribute_Value
join
	AttributeType t
on
	t.AttributeType_ID = a.GenericAttribute_Type
where
	t.AttributeType_Name = 'State'
option (
	maxrecursion 0
);
		
WITH gen(num) AS (
	SELECT
		1
	UNION ALL
	SELECT
		num + 1
	FROM
		gen
	WHERE
		num < @rows
)
insert into SomeTableAttribute 
select
	g.num,
	a.GenericAttribute_ID,
	'2008-01-01',
	1
from
	gen g
join
	GenericAttribute a
on
	cast(g.num as varchar(255)) = a.GenericAttribute_Value
join
	AttributeType t
on
	t.AttributeType_ID = a.GenericAttribute_Type
where
	t.AttributeType_Name = 'Value'
option (
	maxrecursion 0
);

-- select * from SomeTableAttribute

go
create view lSomeTableState
as
select
	sta.SomeTable_ID,
	cast(ga.GenericAttribute_Value as char(3)) as State
from
	SomeTableAttribute sta
join
	GenericAttribute ga
on
	ga.GenericAttribute_ID = sta.GenericAttribute_ID
join
	AttributeType at
on
	at.AttributeType_ID = ga.GenericAttribute_Type
where
	sta.SomeTableAttribute_FromDate = (
		select
			max(sub.SomeTableAttribute_FromDate)
		from
			SomeTableAttribute sub
		where
			sub.SomeTable_ID = sta.SomeTable_ID
	)
and
	at.AttributeType_Name = 'State'

-- select * from lSomeTableState

go
create view lSomeTableValue
as
select
	sta.SomeTable_ID,
	cast(ga.GenericAttribute_Value as int) as Value
from
	SomeTableAttribute sta
join
	GenericAttribute ga
on
	ga.GenericAttribute_ID = sta.GenericAttribute_ID
join
	AttributeType at
on
	at.AttributeType_ID = ga.GenericAttribute_Type
where
	sta.SomeTableAttribute_FromDate = (
		select
			max(sub.SomeTableAttribute_FromDate)
		from
			SomeTableAttribute sub
		where
			sub.SomeTable_ID = sta.SomeTable_ID
	)
and
	at.AttributeType_Name = 'Value'

-- select * from lSomeTableValue

go
create view lSomeTable
as
select
	st.SomeTable_ID,
	sta.State,
	stv.Value
from
	SomeTable st
left join
	lSomeTableState sta
on
	sta.SomeTable_ID = st.SomeTable_ID
left join
	lSomeTableValue stv
on
	stv.SomeTable_ID = st.SomeTable_ID

-- select * from lSomeTable

go
-- testing
DBCC FREEPROCCACHE;
DBCC DROPCLEANBUFFERS;

insert into #timers (test, startTime) 
values(
	'Weak typing', 
	GETDATE()
);

select 
	State,
	max(Value)
from
	lSomeTable
group by
	State;

update #timers 
set 
	stopTime = GETDATE() 
where 
	test = 'Weak typing';
go

-- results
select * from #timers;

/*
declare @value_typed int;
declare @value_untyped varchar(255);

set @value_typed = 1000000;
set @value_untyped = '1000000';

select DATALENGTH(@value_typed);
select DATALENGTH(@value_untyped);
*/
